
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <regex.h>

int strreg( const char *str, const char *reg, regmatch_t *m, size_t n )
{ regex_t preg;
  int r;
	if( regcomp(&preg,reg,REG_EXTENDED)!=0 )
		return(-1);
	r=regexec(&preg,str,n,m,0);
	regfree(&preg);
	return(r);
}

int strcasereg( const char *str, const char *reg, regmatch_t *m, size_t n )
{ regex_t preg;
  int r;
	if( regcomp(&preg,reg,REG_EXTENDED|REG_ICASE)!=0 )
		return(-1);
	r=regexec(&preg,str,n,m,0);
	regfree(&preg);
	return(r);
}

char *get_match( const char *str, regmatch_t *m )
{ char *s;
	if( m->rm_so<0 )
		return(NULL);
	if( (s=malloc(m->rm_eo-m->rm_so+1))==NULL )
		return(NULL);
	strncpy(s,str+m->rm_so,m->rm_eo-m->rm_so);
	s[m->rm_eo-m->rm_so]=0;
	return(s);
}

static void add_to_p( char **p, int *size, int *pos, const char *str, int len )
{
	if( (*pos)+len+1>=(*size) )
	{	*size=(*pos)+len+32;
		*p=realloc(*p,*size);	/* !!! FIXME */
	}
	strncpy((*p)+(*pos),str,len);
	(*pos)+=len;
	(*p)[(*pos)]=0;
}

static char *do_match_apply( const char *str, regmatch_t *m, size_t n, const char *subst, int substitute )
{ char *p;
  int size;
  int i,x;
	
	if( m[0].rm_so<0 )
		return(strdup(str));
	p=NULL; size=0;
	i=0;
	if( substitute )
		add_to_p(&p,&size,&i,str,m[0].rm_so);
	while( *subst )
	{	if( *subst=='\\' )
		{	subst++;
			if( *subst=='n' )
			{	add_to_p(&p,&size,&i,"\n",1);
				subst++;
			}
			else if( *subst>='0' && *subst<='9' )
			{	x=atoi(subst);
				while( *subst>='0' && *subst<='9' )
					subst++;
				if( x<n && m[x].rm_so>=0 )
				{	add_to_p(&p,&size,&i,str+m[x].rm_so,
						m[x].rm_eo-m[x].rm_so);
				}
			}
			else
			{	add_to_p(&p,&size,&i,subst,1);
				subst++;
			}
		}
		else
		{	x=0;
			while( subst[x]!=0 && subst[x]!='\\' )
				x++;
			add_to_p(&p,&size,&i,subst,x);
			subst+=x;
		}
	}
	if( substitute )
		add_to_p(&p,&size,&i,str+m[0].rm_eo,strlen(str+m[0].rm_eo));
	return(p);
}

char *substitute( const char *str, regmatch_t *m, size_t n, const char *subst )
{	return(do_match_apply(str,m,n,subst,1));
}

char *match_apply( const char *str, regmatch_t *m, size_t n, const char *subst )
{	return(do_match_apply(str,m,n,subst,0));
}

char *reg_subst( char *str, char *reg, char *subst, char *flags )
{ int (*sreg)(const char *str, const char *reg, regmatch_t *m, size_t n );
  int global;
  regmatch_t m[32];
  int ofs,i;
  char *p,*q;
//printf("[%s|%s|%s]\"%s\"\r\n",reg,subst,flags,str);
	if( strchr(flags,'i') )
		sreg=strcasereg;
	else	sreg=strreg;
	if( strchr(flags,'g') )
		global=1;
	else	global=0;
	
	ofs=0;
	p=strdup(str);
	do
	{	if( sreg(p+ofs,reg,m,32) )
			break;
		for(i=0;i<32;i++)
		{	if( m[i].rm_so>=0 )
			{	 m[i].rm_so+=ofs;
				 m[i].rm_eo+=ofs;
			}
			else	break;
		}
		q=substitute(p,m,32,subst);
		ofs= strlen(q)-(strlen(p)-m[0].rm_eo);
		free(p);
		p=q;
	} while( global );
	return(p);
}

static char *sed_cmd( char *str, char *cmd )
{ char *r,*s,*f;
  int x;
	if( cmd[0]=='s' && cmd[1]=='/' )
	{	r=cmd+2; x=0;
		while( r[x]!=0 && r[x]!='/' )
		{	if( r[x]=='\\' && r[x+1]=='/' )
				x++;
			x++;
		}
		if( r[x]==0 )	return(NULL);
		r[x]=0;
		s=r+x+1; x=0;
		while( s[x]!=0 && s[x]!='/' )
		{	if( s[x]=='\\' && s[x+1]=='/' )
				x++;
			x++;
		}
		if( s[x]==0 )	return(NULL);
		s[x]=0;
		f=s+x+1;
		return(reg_subst(str,r,s,f));
	}
	return(NULL);
}

char *sed_file( char *str, char *filaname )
{ FILE *f;
  char buf[1024];
  char *p,*q;
  	p=str;
	if( (f=fopen(filaname,"r"))!=NULL )
	{	while( fgets(buf,1024,f) )
		{	if( (q=sed_cmd(p,buf))!=NULL )
			{	if( p!=str )	free(p);
				p=q;
			}
		}
		fclose(f);
	}
	if( p==str )
		p=strdup(str);
	return(p);
}

